import os
from multiprocessing import Pool, cpu_count
from time import sleep
import numpy as np

from models import *

# TODO: get the junk columns out of the pandas parsed datasets

ASC_SFIELDS_EYE = {
    'l':[('onset', np.int64),
         ('x_l', np.float64),
         ('y_l', np.float64),
         ('pup_l', np.float64),],
    'r':[('onset', np.int64),
         ('x_r', np.float64),
         ('y_r', np.float64),
         ('pup_r', np.float64)],
    'b':[('onset', np.int64),
         ('x_l', np.float64),
         ('y_l', np.float64),
         ('pup_l', np.float64),
         ('x_r', np.float64),
         ('y_r', np.float64),
         ('pup_r', np.float64)],}
ASC_SFIELDS_VEL = {
    'l':[('vel_x_l', np.float64),
         ('vel_y_l', np.float64),],
    'r':[('vel_x_r', np.float64),
         ('vel_y_r', np.float64),],
    'b':[('vel_x_l', np.float64),
         ('vel_y_l', np.float64),
         ('vel_x_r', np.float64),
         ('vel_y_r', np.float64)],}
ASC_SFIELDS_REZ = [
    ('res_x', np.float64),
    ('res_y', np.float64)]
ASC_SFIELDS_EXTRA = [('junk', object)]
ASC_SFIELDS_IGNORE = ['junk',]
TXT_FIELDS = {
    'LEFT_ACCELLERATION_X': np.float64,
    'LEFT_ACCELLERATION_Y': np.float64,
    'LEFT_GAZE_X': np.float64,
    'LEFT_GAZE_Y': np.float64,
    'LEFT_IN_BLINK': np.int64,
    'LEFT_IN_SACCADE': np.int64,
    'LEFT_PUPIL_SIZE': np.float64,
    'LEFT_VELOCITY_X': np.float64,
    'LEFT_VELOCITY_Y': np.float64,
    'RECORDING_SESSION_LABEL': object,
    'RIGHT_ACCELLERATION_X': np.float64,
    'RIGHT_ACCELLERATION_Y': np.float64,
    'RIGHT_GAZE_X': np.float64,
    'RIGHT_GAZE_Y': np.float64,
    'RIGHT_IN_BLINK': np.int64,
    'RIGHT_IN_SACCADE': np.int64,
    'RIGHT_PUPIL_SIZE': np.float64,
    'RIGHT_VELOCITY_X': np.float64,
    'RIGHT_VELOCITY_Y': np.float64,
    'TIMESTAMP': np.int64,}
TXT_INT_TYPES = [np.int64]
TXT_NAME_MAP = [np.int64]
ASC_EV_LINE_STARTS = [
    'MSG',
    'START',
    'END',
    'EBLINK',
    'ESACC',
    'EFIX',
    'BUTTON',
    'SAMPLES',]
ASC_EFIELDS_EVENT = {
    'MSG':[('name', object),
           ('onset', np.int64),
           ('label', object),
           ('content', object)],
    'START':[('name', object),
             ('onset', np.int64),
             ('eye', object),
             ('types', object)],
    'END':[('name', object),
           ('onset', np.int64),
           ('types', object),
           ('x_res', np.float64),
           ('y_res', np.float64)],
    'EBLINK':[('name', object),
              ('eye', object),
              ('onset', np.int64),
              ('last_onset', np.int64),
              ('duration', np.int64)],
    'ESACC':[('name', object),
             ('eye', object),
             ('onset', np.int64),
             ('last_onset', np.int64),
             ('duration', np.int64),
             ('x_start', np.float64),
             ('y_start', np.float64),
             ('x_end', np.float64),
             ('y_end', np.float64),
             ('vis_angle', np.float64),
             ('peak_velocity', np.int64)],
    'EFIX':[('name', object),
            ('eye', object),
            ('onset', np.int64),
            ('last_onset', np.int64),
            ('duration', np.int64),
            ('x_pos', np.float64),
            ('y_pos', np.float64),
            ('p_size', np.int64)],
    'BUTTON':[('name', object),
              ('onset', np.int64),
              ('b_num', np.int64),
              ('state', np.int64),],}
ASC_EFIELDS_RES = {
    'MSG':[],
    'START':[],
    'END':[],
    'EBLINK':[],
    'ESACC':[('x_res', np.float64),
             ('y_res', np.float64)],
    'EFIX':[('x_res', np.float64),
             ('y_res', np.float64)],
    'BUTTON':[],}
ASC_EV_IGNORE_COLUMNS = {
    'MSG':[],
    'START':[],
    'END':[],
    'EBLINK':[],
    'ESACC':[],
    'EFIX':[],
    'BUTTON':[],}
ASC_IRREG_EVENTS = ['MSG','START','END']
ASC_INT_TYPES = [np.int64]

def load_eyelink_dataset(file_name):
    """ Parses eyelink data to return samples and events.

    For now, we can only parse events from .asc files. If you hand us a .txt,
    we'll parse out the samples, but not the events.

    Parameters
    ----------
    file_name (string)
        The .asc or .txt file you'd like to parse.

    Returns
    -------
    (Samples object, Events object (or None))
    """
    root, ext = os.path.splitext(file_name)
    if ext == '.asc':
        e, s = pandas_dfs_from_asc(file_name)
    elif ext in ['.txt']:
        s = load_tdf(file_name)
        e = None
    else:
        raise ValueError("only .asc and .txt files supported at the moment...")
    reutn s, e

def pandas_df_from_txt(file_path):
    """ Parses samples out of an EyeLink .txt file """
    import pandas as pd
    import io
    # first we'll just grab everything as objects...
    # then we'll get the fields, figure out the dtypes, and do conversions
    # accordingly. It would be nice if the dtypes would work in read_csv, but
    # so far no luck...
    df = pd.read_csv(file_path, sep="\t", index_col="TIMESTAMP", low_memory=False, na_values=["."],)
    fields = [str(x) for x in df.dtypes.keys()]
    dtypes = dict([(d, object) for d in fields if not d in TXT_FIELDS.keys()])
    dtypes.update(dict([(k, v) for k, v in TXT_FIELDS.iteritems() if k in fields]))
    nums = [k for k, v in dtypes.iteritems() if v not in [object]]
    ints = [k for k in nums if dtypes[k] in TXT_INT_TYPES]
    df[nums] = df[nums].convert_objects(convert_numeric=True)
    df[ints] = df[ints].astype(np.int64)
    # rename TIMESTAMP to "onset" for consistency, and make all columns lower
    fields = [f.lower() for f in fields]
    df.columns = fields
    df.index.name = "onset"
    return Samples.from_pd_obj(df)

def pandas_dfs_from_asc(file_path):
    """ Parses samples and events out of an EyeLink .asc file """
    # collect lines for each event type (including samples)
    e_lines = dict([(k,[]) for k in ASC_EV_LINE_STARTS])
    s_lines = []
    with open(file_path, "r") as f:
        for line in f:
            if len(line) == 0:
                continue
            if line[0].isdigit():
                s_lines.append(line)
                continue
            for k in ASC_EV_LINE_STARTS:
                if line.startswith(k):
                    e_lines[k].append(line)
                    break
    # determine column names, dtypes
    if not len(e_lines["SAMPLES"]) > 0:
        raise ValueError("Could not find samples line in .asc file.")
    side, has_vel, has_res = info_from_asc_samples_line(e_lines["SAMPLES"][0])
    samp_dtypes = build_asc_samp_dtypes(side, has_vel, has_res)
    ev_names = [k for k in ASC_EV_LINE_STARTS if not k in ["SAMPLES"]]
    ev_dtypes = dict([(ev_name, build_asc_ev_dtypes(ev_name, side, has_vel, has_res)) for ev_name in ev_names])
    # get a df for the samples
    samp_df = pandas_df_from_lines(s_lines, samp_dtypes, ASC_SFIELDS_IGNORE)
    # handle event types that need to have their lines preprocessed...
    for ev_name in ASC_IRREG_EVENTS:
        if not ev_name in e_lines:
            continue
        e_lines[ev_name] = prep_irreg_asc_event_lines(e_lines[ev_name], ev_name)
    # get a df for each event type
    ev_dfs = dict([(ev_name,
                    pandas_df_from_lines(e_lines[ev_name],
                                         ev_dtypes[ev_name],
                                         ASC_EV_IGNORE_COLUMNS[ev_name]))
                    for ev_name in ev_names if len(e_lines[ev_name]) > 0])

    # TODO add omitting ASC_EV_IGNORE_COLUMNS[ev_name]
    return Samples.from_pd_obj(samp_df), Events.from_dict(ev_dfs)

def pandas_df_from_lines(csv_lines, dtypes, ignore):
    import pandas as pd
    import cStringIO
    c = cStringIO.StringIO("".join(csv_lines))
    fields, dts = zip(*dtypes)
    # use_names = [n for n in fields if not n in ignore]
    df = pd.read_csv(c,
                     delim_whitespace=True,
                     index_col=["onset"],
                     low_memory=False,
                     na_values=["."],
                     names=fields,
                     header=0,
                     error_bad_lines=False,
                     # usecols=use_names,
                     warn_bad_lines=False,)
    nums = [d[0] for d in dtypes if d[1] not in [object] and d[0] not in ['onset']]
    ints = [d[0] for d in dtypes if d[1] in ASC_INT_TYPES and d[0] not in ['onset']]
    df[nums] = df[nums].convert_objects(convert_numeric=True)
    df[ints] = df[ints].astype(np.int64)
    for ig in ignore:
        del df[ig]
    return df

def prep_irreg_asc_event_lines(lines, ev_name):
    """ uses quotes to force annoying events into usable chunks
    use sparingly - not super fast right now
    """
    new_lines = []
    if ev_name == 'MSG':
        # name, onset, label, content
        # easy - just break content into a third, quoted column
        for line in lines:
            l = line.split()
            lab = l[2] if len(l) > 2 else '.'
            cont = ' '.join(l[3:]) if len(l) > 3 else '.'
            nl = '%s\t%s\t"%s"\t"%s"\n' % (l[0], l[1], lab, cont)
            new_lines.append(nl)
    elif ev_name == 'START':
        # name, onset, eye, then one or two types
        for line in lines:
            l = line.split()
            new_lines.append('%s\t%s\t%s\t"%s"\n' % (l[0],l[1],l[2],', '.join(l[3:])))
    elif ev_name == 'END':
        # name, onset, maybe a list of types, 'RES', x_res, y_res
        # we'll take out the "RES" here
        for line in lines:
            l = line.split()
            types = ' '.join(l[2:-3]) if len(l) > 5 else '.'
            x_res = l[-2]
            y_res = l[-1]
            new_lines.append('%s\t%s\t"%s"\t%s\t%s\n' % (l[0], l[1], types, l[-2], l[-1]))
    else:
        new_lines = lines
    return new_lines

def build_asc_samp_dtypes(side, has_vel, has_res):
    dtypes = list(ASC_SFIELDS_EYE[side])
    if has_vel:
        dtypes.extend(ASC_SFIELDS_VEL[side])
    if has_res:
        dtypes.extend(ASC_SFIELDS_REZ)
    dtypes.extend(ASC_SFIELDS_EXTRA)
    return dtypes

def build_asc_ev_dtypes(ev_name, side, has_vel, has_res):
    dtypes = list(ASC_EFIELDS_EVENT.get(ev_name,[]))
    if has_res:
        dtypes.extend(ASC_EFIELDS_RES.get(ev_name,[]))
    return dtypes if dtypes else None

def info_from_asc_samples_line(line_txt):
    """ gets sample info from asc SAMPLE lines

    Parameters
    ----------
    line_txt (string)
        A single line from an EyeLink asc file.

    Returns
    -------
    side (str)
        'l', 'r', or 'b'
    has_velocity (bool)
        True if velocity information is included in samples
    has_resolution (bool)
        True is resolution information is included in samples
    """
    words = line_txt.split()
    # this line contains information on what the sample lines contain
    has_velocity = "VEL" in words
    has_resolution = "RES" in words
    sample_side = 'b'
    if 'LEFT' in words and not 'RIGHT' in words:
        sample_side = 'l'
    elif 'RIGHT' in words and not 'LEFT' in words:
        sample_side = 'r'
    return sample_side, has_velocity, has_resolution



from models import *
import pandas as pd

#-------------------------------------------------------------
# Masking

def find_nested_events(samples, outer, inner):
    """ Returns indices of events in outer that contain events in inner
    
    This is helpful for dealing with EyeLink blink events. Each is embedded
    within a saccade event, and the EyeLink documentation states that data
    within saccades that contain blinks is unreliable. So we use this method
    to find those saccade events.
    
    Parameters
    ----------
    samples (cili Samples)
        The samples object to which the events objects refer.
    outer (cili Events)
        The list of events you'd like to search for ones containing events in
        'inner.'
    inner (cili Events)
        The list of events whose containing events from 'outer' you're trying
        to find.
    """
    # looking for inner events whose onset is at or before outer offset,
    # and whose offset is at or after inner onset.
    # get list of onsets of first samples *after* inner events
    onsets = inner.index.to_series()
    post_onsets = onsets + inner.duration
    # convert to list of positional indices
    max_onset = samples.index[-1]
    last_idxs = post_onsets.apply(lambda x: max(0, samples.index.searchsorted(x, side="right")-1))
    # step back by one positional index to get pos. index of last samples of our events.
    # stupid fix - don't nudge the index back for events whose duration went beyond the samples
    end_safe_evs = post_onsets <= max_onset
    last_idxs[end_safe_evs] = last_idxs[end_safe_evs] - 1
    # get the time indices of the last samples of our events
    last_onsets = last_idxs.apply(lambda x: samples.index[x])
    idxs = outer.apply(has_overlapping_events, axis=1, args=[onsets, last_onsets])
    if len(idxs) == 0:
        return pd.DataFrame()
    return outer[idxs]

def has_overlapping_events(event, onsets, last_onsets):
    """ Searches for onset/last_onset pairs overlapping with the event in 'event.'

    Specifically, searches series last_onsets for rows with onset <= event
    offset, and offset >= event onset.

    Parameters
    ----------
    event (1xN DataFrame)
        The event you're testing for intersection with the onsets/last_onsets.
    onsets (numpy array like)
        Onset indices of potentially intersecting events.
    last_onsets (numpy array like)
        Last indices of the potentially intersecting events.
    """
    matches = last_onsets[(onsets <= event.name+event.duration) & (last_onsets >= event.name)]
    return len(matches) > 0

def get_eyelink_mask_events(samples, events, find_recovery=True, recovery_field="pup_l"):
    """ Finds events from EyeLink data that contain untrustworthy data.

    Per the EyeLink documentation, we return EBLINK events as well as any
    saccade containing a blink event. We also use adjust_eyelink_recov_idxs to
    push the end of these events slightly farther forward than the EyeLink-
    reported endpoints, because they often include data that is clearly bad.

    Parameters
    ----------
    samples (cili Samples)
        The samples in which the events in 'events' occur.
    events (cili Events)
        The events you'd like to search for blinks and bad saccades.
    find_recovery (bool)
        Defaul True. If true, we will use adjust_eyelink_recov_idxs to find
        the proper ends for blink events.
    recovery_field (string)
        The field used in adjust_eyelink_recov_idxs. Should be either left or
        right pupil size.
    """
    be = events.EBLINK.duration.to_frame()
    be = pd.concat([be, find_nested_events(samples, events.ESACC.duration.to_frame(), be)])
    if find_recovery:
        adjust_eyelink_recov_idxs(samples, be, field=recovery_field)
    return be

def get_eyelink_mask_idxs(samples, events, find_recovery=True, recovery_field="pup_l"):
    """ Calls get_eyelink_mask_events, finds indices from 'samples' within the returned events.

    See notes on get_eyelink_mask_events FMI.
    """
    be = get_eyelink_mask_events(samples, events, find_recovery=find_recovery, recovery_field=recovery_field)
    bi = ev_row_idxs(samples, be)
    return bi

def mask_eyelink_blinks(samples, events, mask_fields=["pup_l"], find_recovery=True, recovery_field="pup_l"):
    """ Sets the value of all untrustworthy data points to NaN.

    Per the EyeLink documentation, we include blink events as well as any
    saccade containing a blink event. We also use adjust_eyelink_recov_idxs to
    push the end of these events slightly farther forward than the EyeLink-
    reported endpoints, because they often include data that is clearly bad.

    Parameters
    ----------
    samples (cili Samples)
        The samples in which the events in 'events' occur.
    events (cili Events)
        The events you'd like to search for blinks and bad saccades.
    mask_fields (list of strings)
        The columns you'd like set to NaN for bad event indices.
    find_recovery (bool)
        Defaul True. If true, we will use adjust_eyelink_recov_idxs to find
        the proper ends for blink events.
    recovery_field (string)
        The field used in adjust_eyelink_recov_idxs. Should be either left or
        right pupil size.
    """
    samps = samples.copy(deep=True)
    indices = get_eyelink_mask_idxs(samps, events, find_recovery=find_recovery, recovery_field=recovery_field)
    samps.loc[indices, mask_fields] = float('nan')
    return samps

def mask_zeros(samples, mask_fields=["pup_l"]):
    """ Sets any 0 values in columns in mask_fields to NaN

    Parameters
    ----------
    samples (cili Samples)
        The samples you'd like to search for 0 values.
    mask_fields (list of strings)
        The columns in you'd like to search for 0 values.
    """
    samps = samples.copy(deep=True)
    for f in mask_fields:
        samps[samps[f] == 0] = float("nan")
    return samps

def interp_zeros(samples, interp_fields=["pup_l"]):
    """ Replace 0s in 'samples' with linearly interpolated data.

    Parameters
    ----------
    samples (cili Samples)
        The samples in which you'd like to replace 0s
    interp_fields (list of strings)
        The column names from samples in which you'd like to replace 0s.
    """
    samps = mask_zeros(samples, mask_fields=interp_fields)
    samps = samps.interpolate(method="linear", axis=0, inplace=False)
    # since interpolate doesn't handle the start/finish, bfill the ffill to
    # take care of NaN's at the start/finish samps.
    samps.fillna(method="bfill", inplace=True)
    samps.fillna(method="ffill", inplace=True)
    return samps

def interp_eyelink_blinks(samples, events, find_recovery=True, interp_fields=["pup_l"], recovery_field="pup_l"):
    """ Replaces the value of all untrustworthy data points linearly interpolated data.

    Per the EyeLink documentation, we include blink events as well as any
    saccade containing a blink event. We also use adjust_eyelink_recov_idxs to
    push the end of these events slightly farther forward than the EyeLink-
    reported endpoints, because they often include data that is clearly bad.

    Parameters
    ----------
    samples (cili Samples)
        The samples in which the events in 'events' occur.
    events (cili Events)
        The events you'd like to search for blinks and bad saccades.
    find_recovery (bool)
        Defaul True. If true, we will use adjust_eyelink_recov_idxs to find
        the proper ends for blink events.
    interp_fields (list of strings)
        The columns in which we should interpolate data.
    recovery_field (string)
        The field used in adjust_eyelink_recov_idxs. Should be either left or
        right pupil size.
    """
    samps = mask_eyelink_blinks(samples, events, mask_fields=interp_fields, find_recovery=find_recovery, recovery_field=recovery_field)
    # inplace=True causes a crash, so for now...
    # fixed by #6284 ; will be in 0.14 release of pandas
    samps = samps.interpolate(method="linear", axis=0, inplace=False)
    return samps

def ev_row_idxs(samples, events):
    """ Returns the indices in 'samples' contained in events from 'events.'

    Parameters
    ----------
    samples (cili Samples)
        The samples you'd like to pull indices from.
    events (cili Events)
        The events whose indices you'd like to pull from 'samples.'
    """
    import numpy as np
    idxs = []
    for idx, dur in events.duration.iteritems():
        idxs.extend(range(idx, int(idx+dur)))
    idxs = np.unique(idxs)
    idxs = np.intersect1d(idxs, samples.index.tolist())
    return idxs

def adjust_eyelink_recov_idxs(samples, events, z_thresh=.1, field="pup_l", window=1000, kernel_size=100):
    """ Extends event endpoint until the z-scored derivative of 'field's timecourse drops below thresh

    We will try to extend *every* event passed in.

    Parameters
    ----------
    samples (list of dicts)
        A Samples object
    events (list of dicts)
        An Events object
    z_thresh (float)
        The threshold below which the z-score of the timecourse's gradient
        must fall before we'll consider the event over.
    field (string)
        The field to use.
    window (int)
        The number of indices you'll search through for z-threshold
    kernel_size (int)
        The number of indices we'll average together at each measurement. So
        what you really get with this method is the index of the first voxel
        whose gradient value, when averaged together with that of the
        n='kernel' indices after it, then z-scored, is below the given z
        threshold.
    """
    import numpy as np
    # use pandas to take rolling mean. pandas' kernel looks backwards, so we need to pull a reverse...
    dfs = np.gradient(samples[field].values)
    reversed_dfs = dfs[::-1]
    reversed_dfs_ravg = np.array(pd.rolling_mean(pd.Series(reversed_dfs),window=kernel_size, min_periods=1))
    dfs_ravg = reversed_dfs_ravg[::-1]
    dfs_ravg = np.abs((dfs_ravg-np.mean(dfs_ravg))/np.std(dfs_ravg))
    samp_count = len(samples)
    # search for drop beneath z_thresh after end index
    new_durs = []
    for idx, dur in events.duration.iteritems():
        try:
            s_pos = samples.index.get_loc(idx + dur)  - 1
            e_pos = samples.index[min(s_pos+window, samp_count-1)]
        except Exception, e:
            # can't do much about that
            s_pos = e_pos = 0
        if s_pos == e_pos:
            new_durs.append(dur)
            continue
        e_dpos = np.argmax(dfs_ravg[s_pos:e_pos] < z_thresh) # 0 if not found
        new_end = samples.index[min(s_pos + e_dpos, samp_count-1)]
        new_durs.append(new_end - idx)
    events.duration = new_durs

#-------------------------------------------------------------
# Filters

def butterworth_series(samples, fields=["pup_l"], filt_order=5, cutoff_freq=.01, inplace=False):
    """ Applies a butterworth filter to the given fields

    See documentation on scipy's butter method FMI.
    """
    # TODO: This is pretty limited right now - you'll have to tune filt_order
    # and cutoff_freq manually. In the future, it would be nice to use
    # signal.buttord (heh) to let people adjust in terms of dB loss and
    # attenuation.
    import scipy.signal as signal
    from numpy import array
    samps = samples if inplace else samples.copy(deep=True)
    B, A = signal.butter(filt_order, cutoff_freq, output="BA")
    samps[fields] = samps[fields].apply(lambda x: signal.filtfilt(B,A,x), axis=0)
    return samps




import pandas as pd
import pandas.io.pytables as pt
from pandas.compat import u_safe as u, string_types, isidentifier

class SaveMixin(object):
    """ Bakes in some save settings for NDFrame subclasses
    
    You can still use the pandas methods, but for quick saving and loading
    this mixin provides some setting you might want to reuse.
    """
    def __init__(self, *args, **kwargs):
        super(SaveMixin, self).__init__(*args, **kwargs)

    def save(self, save_path):
        # explore slower but more flexible "table" later
        initialize_hdf5()
        self.to_hdf(save_path, "obj", format="fixed")

    @classmethod
    def load_saved(cls, save_path):
        from pandas import read_hdf
        initialize_hdf5()
        obj = read_hdf(save_path, "obj")
        obj2 = cls.from_pd_obj(obj)
        return obj2

    @classmethod
    def from_pd_obj(cls, pd_obj):
        return cls(pd_obj._data.copy()).__finalize__(pd_obj)

class Samples(SaveMixin, pd.DataFrame):
    """Pandas DataFrame subclas for representing eye tracking timeseries data.

    Indexes may be hierarchical.
    """
    def __init__(self, *args, **kwargs):
        super(Samples, self).__init__(*args, **kwargs)
        
class Events(object):
    """Pandas Panel-like object that gives you access to DataFrames via standard accessors.

    Also implements save and load_saved. And init form dict. Otherwise, it's
    not much like a Panel where it counts.

    One advantage to avoiding Panel is that this lets us bundle all of our
    event DataFrames into one object without loding the column dtypes during
    merge.

    Right now, the best way way to make one of these is to use Events.from_dict().
    """
    def __init__(self, *args, **kwargs):
        super(Events, self).__init__(*args, **kwargs)
        self.dframes = {}

    def save(self, save_path):
        s = pt.HDFStore(save_path)
        for k in self.dframes.keys():
            s[k] = self.dframes[k]
        s.close()

    @classmethod
    def load_saved(cls, save_path):
        obj = cls()
        s = pt.HDFStore(save_path)
        obj.dframes = dict([(k[1:],s[k]) for k in s.keys()])
        s.close()
        return obj

    @classmethod
    def from_dict(cls, the_d):
        """ Returns an Events instance containing the given DataFrames
        
        Parameters
        ----------
        the_d (dict)
            A dictionary with event names for keys, and DataFrames for values.
        """
        obj = cls()
        obj.dframes = the_d
        return obj

    @classmethod
    def from_list_of_dicts(cls, events_list):
        """ Returns an Events instance built using the given dicts.

        Parameters
        ----------
        events_list (list of dictionaries, or path to pickled file)
            This can either be an actual list, or the location of a pickled
            list. We'll group the events by 'name' into DataFrames, then pass
            those to from_dict. Note that only field that appear in every
            dictionary with a given name will make it through to the
            corresponding DataFrame.
        """
        import os
        from cPickle import load as pload
        if isinstance(events_list, str):
            if not os.path.isfile(events_list):
                raise ValueError("Could not find file %s" % events_list)
            evl = pload(events_list)
        else:
            evl = events_list
        evd = {}
        for ev in evl:
            if not "name" in ev:
                continue
            if not ev["name"] in evd:
                evd[ev["name"]] = [ev]
            else:
                evd[ev["name"]].append(ev)
        for k in evd.keys():
            evd[k] = pd.DataFrame(evd[k])
        return cls.from_dict(evd)

    def _local_dir(self):
        """ add the string-like attributes from the info_axis """
        return [c for c in self.dframes.keys()
            if isinstance(c, string_types) and isidentifier(c)]

    def __dir__(self):
        """
        Provide method name lookup and completion
        Only provide 'public' methods
        """
        return list(sorted(list(set(dir(type(self)) + self._local_dir()))))

    def __getattr__(self, name):
        """Similar to Panel behavior, but with filtered accessors.

        Rows will be filtered by any(notnull(columns)), then columns filtered
        by any(notnull(rows)). This way, you should quickly be able to get
        ahold of and inspect the set of events and values you're interested
        in.
        """
        if name in self.dframes:
            return self.dframes[name]
        raise AttributeError("'%s' object has no attribute '%s'" %
                             (type(self).__name__, name))

def initialize_hdf5():
    pt._TYPE_MAP.update({Events:u('wide'), Samples:u('frame'),})
    pt._AXES_MAP.update({Events:[1, 2], Samples:[0],})
